/* conf.c
    Routines to set configuration options.
    Copyright (C) 1999 Eric Gillespie, Jr.

    This file is part of dpsftplib.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,  as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>

#include "dpsftplib.h"
#include "conf.h"

struct configdata Conf;

static gboolean
SetString(gchar **dest, gchar *src)
{
	if (*dest)
		g_free(*dest);

	*dest = g_strdup(src);
	if (!*dest) {
		ftperrno = FTP_SYS;
		return FALSE;
	}

	return TRUE;
}

gboolean
ftpConfSetAnonPass(gchar *val)
{
	return SetString(&Conf.anonpass, val);
}

gboolean
ftpConfSetQueueFile(gchar *val)
{
	return SetString(&Conf.queuefile, val);
}

void
ftpConfSetReuseConnection(gboolean val)
{
	Conf.reuseConnection = val;
}

void
ftpConfSetQueuebytes(gint val)
{
	Conf.queuebytes = val;
}

static gboolean
GetValue(GScanner *scanner)
{
	GTokenType token;
	gchar *var;
	gint pos;

	g_return_val_if_fail(scanner != NULL, FALSE);

	var = g_strdup(scanner->value.v_string);

	pos = g_scanner_cur_position(scanner) + 1;
	token = g_scanner_get_next_token(scanner);
	if (token != G_TOKEN_EQUAL_SIGN) {
		g_scanner_error(scanner,
				"'=' expected at column %d", pos);
		g_free(var);
		return FALSE;
	}

	if (g_strncasecmp(var, "anonpass", 8) == 0) {
		g_free(var);
		pos = g_scanner_cur_position(scanner) + 1;
		token = g_scanner_get_next_token(scanner);
		if (token != G_TOKEN_STRING) {
			g_scanner_error(scanner,
					"String expected at column %d,"
					" try putting it in quotes", pos);
			return FALSE;
		}
		Conf.anonpass = ftpParseEnv(scanner->value.v_string);

	} else if (g_strncasecmp(var, "reuseConnection", 19) == 0) {
		g_free(var);
		pos = g_scanner_cur_position(scanner) + 1;
		token = g_scanner_get_next_token(scanner);
		if (token != G_TOKEN_SYMBOL) {
			g_scanner_error(scanner,
					"TRUE or FALSE expected at"
					" column %d", pos);
			return FALSE;
		}
		Conf.reuseConnection = (gboolean) scanner->value.v_symbol;

	} else if (g_strncasecmp(var, "queuefile", 9) == 0) {
		g_free(var);
		pos = g_scanner_cur_position(scanner) + 1;
		token = g_scanner_get_next_token(scanner);
		if (token != G_TOKEN_STRING) {
			g_scanner_error(scanner,
					"String expected at column %d,"
					" try putting it in quotes", pos);
			return FALSE;
		}
		Conf.queuefile = ftpParseEnv(scanner->value.v_string);

	} else if (g_strncasecmp(var, "queuebytes", 10) == 0) {
		g_free(var);
		pos = g_scanner_cur_position(scanner) + 1;
		token = g_scanner_get_next_token(scanner);
		if (token != G_TOKEN_INT) {
			g_scanner_error(scanner,
					"Integer expected at column %d",
					pos);
			return FALSE;
		}
		Conf.queuebytes = scanner->value.v_int;

	} else {
		g_free(var);
		g_scanner_error(scanner, "Unknown identifer %s", var);
		return FALSE;
	}

	return TRUE;
}

gboolean
ftpLoadConf(gchar *fn)
{
	GScanner *scanner;
	GTokenType token;
	gint fd, pos;
	gboolean error;

	g_return_val_if_fail(fn != NULL, FALSE);

	fd = open(fn, O_RDONLY);
	if (fd < 0) {
		ftperrno = FTP_SYS;
		return FALSE;
	}

	scanner = g_scanner_new(NULL);
	if (!scanner) {
		close(fd);
		return FALSE;
	}
	g_scanner_input_file(scanner, fd);
	scanner->input_name = fn;

	g_scanner_set_scope(scanner, 42);
	g_scanner_freeze_symbol_table(scanner);
	g_scanner_scope_add_symbol(scanner, 42, "TRUE", (gpointer) TRUE);
	g_scanner_scope_add_symbol(scanner, 42, "FALSE", (gpointer) FALSE);
	g_scanner_thaw_symbol_table(scanner);

	while (g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) {
		pos = g_scanner_cur_position(scanner) + 1;
		token = g_scanner_get_next_token(scanner);
		error = FALSE;
		if (token == G_TOKEN_IDENTIFIER) {
			if (!GetValue(scanner))
				error = TRUE;
		} else {
			g_scanner_error(scanner,
					"Identifier expected at column %d",
					pos);
			error = TRUE;
		}
		if (error) {
			g_scanner_destroy(scanner);
			close(fd);
			return FALSE;
		}
	}

	g_scanner_destroy(scanner);
	close(fd);

        return TRUE;
}

gboolean
ftpSaveConf(gchar *fn)
{
	gint fd;
	FILE *file;

	g_return_val_if_fail(fn != NULL, FALSE);

	fd = creat(fn, S_IRUSR | S_IWUSR);
	if (fd < 0)
		return FALSE;

	file = fdopen(fd, "w");
	if (!file) {
		close(fd);
		return FALSE;
	}

	fprintf(file, "anonpass=%s\n", Conf.anonpass);
	fprintf(file, "queuefile=%s\n", Conf.queuefile);
	fprintf(file, "queuebytes=%d\n", Conf.queuebytes);
	fclose(file);

	return TRUE;
}
