/* ftpfile.c
    Routines to manage the ftpFile data type.
    Copyright (C) 1999 Eric Gillespie, Jr.

    This file is part of dpsftplib.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,  as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>

#include "dpsftplib.h"
#include "ftpfile.h"

ftpFile *
ftpFileNew(const gchar *name, const gchar *dir, guint32 size,
	   const gchar *date, gint type, const gchar *mode, const gchar *user,
	   const gchar *group)
{
	ftpFile *file;

	g_return_val_if_fail(name != NULL, NULL);
	g_return_val_if_fail(dir != NULL, NULL);
	g_return_val_if_fail((type == 'A') || (type == 'I'), NULL);

	ftperrno = FTP_SYS;

	file = g_malloc(sizeof(ftpFile));
	if (!file)
		return NULL;
	else
		file->name = file->dir = file->date =
			file->mode = file->user = file->group = NULL;

	file->name = g_strdup(name);
	if (!file->name) {
		ftpFileDelete(file);
		return NULL;
	}

	file->dir = g_strdup(dir);
	if (!file->dir) {
		ftpFileDelete(file);
		return NULL;
	}

	if (date) {
		file->date = g_strdup(date);
		if (!file->date) {
			ftpFileDelete(file);
			return NULL;
		}
	}

	if (mode) {
		file->mode = g_strdup(mode);
		if (!file->mode) {
			ftpFileDelete(file);
			return NULL;
		}
	}

	if (user) {
		file->user = g_strdup(user);
		if (!file->user) {
			ftpFileDelete(file);
			return NULL;
		}
	}

	if (group) {
		file->group = g_strdup(group);
		if (!file->group) {
			ftpFileDelete(file);
			return NULL;
		}
	}

	file->size = size;
	file->type = type;

	return file;
}

void
ftpFileDelete(ftpFile *file)
{
	g_return_if_fail(file != NULL);

	if (file->name)
		g_free(file->name);

	if (file->date)
		g_free(file->date);

	if (file->dir)
		g_free(file->dir);

	if (file->mode)
		g_free(file->mode);

	if (file->user)
		g_free(file->user);

	if (file->group)
		g_free(file->group);

	g_free(file);
}

gint
ftpFileGetType(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, 0);

	return file->type;
}

void
ftpFileSetType(ftpFile *file, gint type)
{
	g_return_if_fail(file != NULL);
	g_return_if_fail((type == 'A') || (type == 'I'));

	file->type = type;
}

const gchar *
ftpFileGetName(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, NULL);

	return file->name;
}

const gchar *
ftpFileGetDir(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, NULL);

	return file->dir;
}

guint32
ftpFileGetSize(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, -1);

	return file->size;
}

const gchar *
ftpFileGetMode(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, NULL);

	return file->mode;
}

const gchar *
ftpFileGetUser(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, NULL);

	return file->user;
}

const gchar *
ftpFileGetGroup(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, NULL);

	return file->group;
}

const gchar *
ftpFileGetDate(const ftpFile *file)
{
	g_return_val_if_fail(file != NULL, NULL);

	return file->date;
}
