/* ftpsite.c
    Routines to manage the ftpSite data type.
    Copyright (C) 1999 Eric Gillespie, Jr.

    This file is part of dpsftplib.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,  as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <errno.h>
#include <pthread.h>

#include "dpsftplib.h"
#include "ftpsite.h"

#define LENGTH 513

ftpSite *
ftpSiteNew(gchar *hostname, gint port, gchar *login, gchar *pass,
	   ftpLogCallback logfunc, gpointer command, gpointer reply,
	   gpointer logdata, gint type)
{
	ftpSite *site;

	g_return_val_if_fail(hostname != NULL, NULL);
	g_return_val_if_fail(login != NULL, NULL);
	g_return_val_if_fail(pass != NULL, NULL);
	g_return_val_if_fail((type == 'I') || (type == 'A'), NULL);

	ftperrno = FTP_SYS;

	site = g_malloc(sizeof(ftpSite));
	if (!site)
		return NULL;

	site->hostname = site->login = site->pass = site->pwd = NULL;
	site->replybuffer = NULL;
	site->transfer = NULL;

	site->hostname = g_strdup(hostname);
	if (!site->hostname) {
		ftpSiteDelete(site);
		return NULL;
	}

	site->login = g_strdup(login);
	if (!site->login) {
		ftpSiteDelete(site);
		return NULL;
	}

	site->pass = g_strdup(pass);
	if (!site->pass) {
		ftpSiteDelete(site);
		return NULL;
	}

	site->replybuffer = g_malloc(LENGTH);
	if (!site->replybuffer) {
		ftpSiteDelete(site);
		return NULL;
	}
	site->replybuffer[0] = '\0';

	site->port = port;
	site->logfunc = logfunc;
	site->command = command;
	site->reply = reply;
	site->logdata = logdata;
	site->type = type;

	site->connected = FALSE;

	site->dirlist = site->filelist = NULL;

	pthread_mutex_init(&site->mutex, NULL);

	return site;
}

void
ftpSiteDelete(ftpSite *site)
{
	g_return_if_fail(site != NULL);

	if (site->connected)
		ftpClose(site);

	if (site->hostname)
		g_free(site->hostname);

	if (site->login)
		g_free(site->login);

	if (site->pass)
		g_free(site->pass);

	if (site->pwd)
		g_free(site->pwd);

	if (site->replybuffer)
		g_free(site->replybuffer);

	g_list_foreach(site->dirlist, (GFunc) ftpDirectoryDelete, NULL);
	g_list_free(site->dirlist);

	g_free(site);
}

ftpSite *
ftpSiteCopy(ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return ftpSiteNew(site->hostname, site->port, site->login, site->pass,
			  site->logfunc, site->command, site->reply,
			  site->logdata, site->type);
}

void
ftpSiteSetLogCallback(ftpSite *site, ftpLogCallback logfunc)
{
	g_return_if_fail(site != NULL);

	site->logfunc = logfunc;
}

ftpLogCallback
ftpSiteGetLogCallback(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->logfunc;
}

void
ftpSiteSetLogCommandData(ftpSite *site, gpointer data)
{
	g_return_if_fail(site != NULL);

	site->command = data;
}

gpointer
ftpSiteGetLogCommandData(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->command;
}

void
ftpSiteSetLogReplyData(ftpSite *site, gpointer data)
{
	g_return_if_fail(site != NULL);

	site->reply = data;
}

gpointer
ftpSiteGetLogReplyData(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->reply;
}

void
ftpSiteSetLogData(ftpSite *site, gpointer data)
{
	g_return_if_fail(site != NULL);

	site->logdata = data;
}

gpointer
ftpSiteGetLogData(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->logdata;
}

void
ftpSiteSetLog(ftpSite *site, ftpLogCallback logfunc, gpointer command,
	      gpointer reply, gpointer logdata)
{
	g_return_if_fail(site != NULL);

	ftpSiteSetLogCallback(site, logfunc);
	ftpSiteSetLogCommandData(site, command);
	ftpSiteSetLogReplyData(site, reply);
	ftpSiteSetLogData(site, logdata);
}

gint
ftpSiteGetType(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, 0);

	return site->type;
}

void
ftpSiteSetType(ftpSite *site, gint type)
{
	g_return_if_fail(site != NULL);

	site->type = type;
}

gboolean
ftpSiteConnected(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, FALSE);

	return site->connected;
}

const gchar *
ftpSiteGetHostname(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->hostname;
}

void
ftpSiteSetHostname(ftpSite *site, gchar *hostname)
{
	g_return_if_fail(site != NULL);
	g_return_if_fail(hostname != NULL);

	site->hostname = g_strdup(hostname);
}

gint
ftpSiteGetPort(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, -1);

	return site->port;
}

void
ftpSiteSetPort(ftpSite *site, gint port)
{
	g_return_if_fail(site != NULL);

	site->port = port;
}

const gchar *
ftpSiteGetLogin(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->login;
}

void
ftpSiteSetLogin(ftpSite *site, gchar *login)
{
	g_return_if_fail(site != NULL);
	g_return_if_fail(login != NULL);

	site->login = g_strdup(login);
}

const gchar *
ftpSiteGetPass(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->pass;
}

void
ftpSiteSetPass(ftpSite *site, gchar *pass)
{
	g_return_if_fail(site != NULL);

	site->pass = g_strdup(pass);
}

const gchar *
ftpSiteGetReplyBuffer(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->replybuffer;
}

const gchar *
ftpSiteGetPwd(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, NULL);

	return site->pwd;
}

gboolean
ftpSiteTransferring(const ftpSite *site)
{
	g_return_val_if_fail(site != NULL, FALSE);

	return (site->transfer == NULL ? FALSE : TRUE);
}

gint
ftpSiteLock(ftpSite *site)
{
	g_return_val_if_fail(site != NULL, EINVAL);

	return pthread_mutex_trylock(&site->mutex);
}

gint
ftpSiteUnlock(ftpSite *site)
{
	g_return_val_if_fail(site != NULL, EINVAL);


	return pthread_mutex_unlock(&site->mutex);
}
