/* main.c
    Main routines for dpsftplib.
    Copyright (C) 1999 Eric Gillespie, Jr.

    This file is part of dpsftplib.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,  as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <glib.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "dpsftplib.h"
#include "conf.h"
#include "paths.h"
#include "ftpdirectory.h"
#include "ftpfile.h"
#include "ftpsite.h"
#include "queue.h"

#define LENGTH 513

#define LIST_ARGS "-aL"
gint errno;


static GList *localdirlist;
static pthread_mutex_t localdirmutex = PTHREAD_MUTEX_INITIALIZER;
static GList *localfilelist;
static pthread_mutex_t localfilemutex = PTHREAD_MUTEX_INITIALIZER;

const gchar *ftp_errlist[] = {
	/* Connection errors */
	[FTP_LOSTCONN] = "Lost the connection",
	[FTP_SOCKFAIL] = "Failed to create socket",
	[FTP_FILESOCKFAIL] = "Failed to open stream for socket",
	[FTP_CONNECTED] = "Already connected",
	[FTP_NOTCONNECTED] = "Not connected",

	/* Command errors */
	[FTP_USERFAIL] = "Login command USER failed",
	[FTP_PASSFAIL] = "Login command PASS failed",
	[FTP_TYPEFAIL] = "Failed to set transfer type",
	[FTP_RLISTFAIL] = "Failed to get remote file listing",
	[FTP_LLISTFAIL] = "Failed to get local file listing",
	[FTP_RCHDIRFAIL] = "Failed to change remote directory",
	[FTP_LCHDIRFAIL] = "Failed to change local directory",
	[FTP_NOOPFAIL] = "Keepalive command NOOP failed",

	/* Transfer errors */
	[FTP_TRANSFERRING] = "Already transferring",
	[FTP_NOTTRANSFERRING] = "Not transferring",
	[FTP_TRANSFERFAIL] = "Failed to open file transfer",
	[FTP_NOTDOWN] = "Transfer is not a download",
	[FTP_NOTUP] = "Transfer is not an upload",
	[FTP_TLOSTCONN] = "Lost transfer connection",
	[FTP_RESTFAIL] = "Failed to resume transfer",
	[FTP_ABORTFAIL] = "Transfer abort failed, transfer continuing",

	/* Queue errors */
	[FTP_QGOING] = "Queue is already running",
	[FTP_QNOTQUEUED] = "Item not found in queue",
	[FTP_QFIRST] = "Item is already first in queue",
	[FTP_QLAST] = "Item is already last in queue",

	/* Miscellaneous errors */
	[FTP_SYS] = "System error, system errno says"
};

const gint ftp_nerr = sizeof(ftp_errlist) / sizeof(ftp_errlist[0]);

#ifdef HAVE_PTHREADS
pthread_key_t ftperrno_key;
#else
ftpError ftperrno;
#endif

static gchar SystemConfFn[] = SYSCONFDIR "/system.dpsftplibrc";
static gchar DefaultConfFn[] = "$HOME/.dpsftplibrc";

#ifdef HAVE_PTHREADS
void
ftpThreadInit(void)
{
	ftpError *err;

	err = g_malloc(sizeof(ftpError));
	pthread_setspecific(ftperrno_key, err);
}
#endif

gboolean
ftpInit(gchar *config)
{
	gchar *fn;

	localdirlist = localfilelist = NULL;

#ifdef HAVE_PTHREADS
	pthread_key_create(&ftperrno_key, g_free);

	ftpThreadInit();
#endif

	if (!ftpLoadConf(SystemConfFn))
		return FALSE;

	fn = ftpParseEnv(DefaultConfFn);
	ftpLoadConf(fn);
	g_free(fn);

	if (config) {
		if (!ftpLoadConf(config))
			return FALSE;
	}

	if (!ftpQueueInit())
		return FALSE;

	return TRUE;
}

gboolean
ftpShutdown(void)
{
	if (!ftpQueueShutdown())
		return FALSE;

	g_list_foreach(localdirlist, (GFunc) ftpDirectoryDelete, NULL);
	g_list_free(localdirlist);

	g_list_foreach(localfilelist, (GFunc) ftpFileDelete, NULL);
	g_list_free(localfilelist);

	g_free(Conf.anonpass);
	g_free(Conf.queuefile);

	return TRUE;
}

gint
ftpCommand(const ftpSite *site, gchar *format, ...)
{
	gint r;
	gchar *cmd;
	va_list ap;

	g_return_val_if_fail(site != NULL, 0);
	g_return_val_if_fail(format != NULL, 0);

	if (!ftpSiteConnected(site)) {
		ftperrno = FTP_NOTCONNECTED;
		return 0;
	}

	va_start(ap, format);
	cmd = g_strdup_vprintf(format, ap);
	va_end(ap);

	if (site->logfunc) {
		if (strncmp(cmd, "PASS", 4) == 0)
			site->logfunc(site->logdata, site->command,
				      "PASS XXXX\r\n");
		else
			site->logfunc(site->logdata, site->command, cmd);
	}

	r = fprintf(site->cout, "%s", cmd);
	g_free(cmd);

	return r;
}

/* ONLY use this function to get replies from commands you've issued
   by calling ftpCommand(). Any other use could break the library. */
const gchar *
ftpGetReply(ftpSite *site)
{
	gchar *buf;

	g_return_val_if_fail(site != NULL, NULL);

	if (!ftpSiteConnected(site)) {
		ftperrno = FTP_NOTCONNECTED;
		return NULL;
	}

	buf = site->replybuffer;

        do {
		if (!fgets(buf, LENGTH, site->cin))
			return NULL;
		if (site->logfunc)
			site->logfunc(site->logdata, site->reply, "%s", buf);
        } while (buf[3] == '-' || buf[0] == ' ');

        return buf;
}

gboolean
ftpOpen(ftpSite *site)
{
	gint socket;

	g_return_val_if_fail(site != NULL, FALSE);

	if (site->connected) {
		ftperrno = FTP_CONNECTED;
		return FALSE;
	}
	
	socket = OpenSocket(site->hostname, site->port);
	if (socket < 0) {
		ftperrno = FTP_SOCKFAIL;
		return FALSE;
	}

	site->cout = FileSocket(socket, "wt");
	if (!site->cout) {
		ftperrno = FTP_FILESOCKFAIL;
		return FALSE;
	}

	site->cin = FileSocket(socket, "rt");
	if (!site->cin) {
		ftperrno = FTP_FILESOCKFAIL;
		fclose(site->cout);
		return FALSE;
	}

	site->connected = TRUE;
	ftpGetReply(site);

	if (!ftpCommand(site, "USER %s\r\n", site->login)) {
		ftperrno = FTP_LOSTCONN;
		fclose(site->cout);
		site->connected = FALSE;
		return FALSE;
	}

	if (ftpGetReply(site)[0] != '3') {
		ftperrno = FTP_USERFAIL;
		fclose(site->cout);
		site->connected = FALSE;
                return FALSE;
	}

	if (!ftpCommand(site, "PASS %s\r\n", site->pass)) {
		ftperrno = FTP_LOSTCONN;
		fclose(site->cout);
		site->connected = FALSE;
		return FALSE;
	}

        if (ftpGetReply(site)[0] != '2') {
		ftperrno = FTP_PASSFAIL;
		fclose(site->cout);
		site->connected = FALSE;
                return FALSE;
	}

	if (!ftpPwd(site)) {
		fclose(site->cout);
		site->connected = FALSE;
		return FALSE;
	}

	/* RFC 959 states that ASCII is the default transfer mode. */
	site->type = 'A';

	return TRUE;
}

gboolean
ftpClose(ftpSite *site)
{
	g_return_val_if_fail(site != NULL, FALSE);

	if (!site->connected) {
		ftperrno = FTP_NOTCONNECTED;
		return FALSE;
	}

        if (!ftpCommand(site, "QUIT\r\n")) {
		ftperrno = FTP_LOSTCONN;
		return FALSE;
	}

	ftpGetReply(site);
	site->connected = FALSE;

	fclose(site->cout);

	return TRUE;
}

gboolean
ftpOpenData(ftpSite *site, gchar *mode)
{
	gint a1, a2, a3, a4, p1, p2, port, socket;
	gchar *ip;

	g_return_val_if_fail(site != NULL, FALSE);
	g_return_val_if_fail(mode != NULL, FALSE);

	if (!ftpCommand(site, "PASV\r\n")) {
		ftperrno = FTP_LOSTCONN;
		return FALSE;
	}

	sscanf(ftpGetReply(site), "%*d %*s %*s %*s %*c%d,%d,%d,%d,%d,%d%*c",
	       &a1, &a2, &a3, &a4, &p1, &p2);
	ip = g_strdup_printf("%d.%d.%d.%d", a1, a2, a3, a4);
	port = ( (p1 << 8) | p2 );

	socket = OpenSocket(ip, port);
	if (socket < 0) {
		ftperrno = FTP_SOCKFAIL;
		return FALSE;
	}
	site->data = FileSocket(socket, mode);
	if (!site->data) {
		ftperrno = FTP_FILESOCKFAIL;
		return FALSE;
	}

	g_free(ip);

	return TRUE;
}

static ftpFile *
ParseLs(gchar *buf, GList **filelist, const gchar *dir)
{
	gchar *size, *month, *day, *year, *date, *fn, *tmp;
	gchar *mode, *user, *group;
	ftpFile *file;
	gboolean islink = FALSE;

	mode = strtok(buf, " ");
	strtok(NULL, " ");
	user = strtok(NULL, " ");
	group = strtok(NULL, " ");
	size = strtok(NULL, " ");
	month = strtok(NULL, " ");
	day = strtok(NULL, " ");
	year = strtok(NULL, " ");

	fflush(stdout);
	/* we can't use strtok because filenames may have spaces */
	fn = year + strlen(year) + 1;
	g_strchomp(fn);
	if (strcmp(fn, ".") == 0 || strcmp(fn, "..") == 0)
		return NULL;

	if (buf[0] == 'l') {
		if ((tmp = strstr(fn, "->"))) {
			islink = TRUE;
			(tmp - 1)[0] = '\0';
		}
	}

	date = g_strdup_printf("%s %s %s", month, day, year);

	file = ftpFileNew(fn, dir, strtol(size, NULL, 10), date, 'I',
			  mode, user, group);
	if (!file)
		return NULL;

	g_free(date);

	if (buf[0] == 'd')
		file->type = 'd';
	else {
		if (islink)
			file->type = 'l';
		else
			/* eventually we'll determine its type */
			file->type = 'I';

		*filelist = g_list_append(*filelist, file);
	}

	return file;
}

ftpFile *
ftpGetFileInfoRemote(ftpSite *site, const gchar *dir, const gchar *name)
{
	gchar *buf, *pwd;
	GList *filelist;
	ftpFile *file, *queuefile = NULL;

	g_return_val_if_fail(site != NULL, NULL);
	g_return_val_if_fail(name != NULL, NULL);

	pwd = g_strdup(site->pwd);

	if (!dir)
		dir = pwd;

	for (filelist = site->filelist; filelist; filelist = filelist->next) {
		file = filelist->data;
		if (strcmp(file->dir, dir) == 0
		    && strcmp(file->name, name) == 0) {
			if (!file->date || !file->mode || !file->user
			    || !file->group) {
				queuefile = file;
			}
			else {
				g_free(pwd);
				return file;
			}
		}
	}

	if (!ftpChdir(site, dir)) {
		g_free(pwd);
		return NULL;
	}

	if (!ftpCommand(site, "SIZE %s\r\n", name)) {
		g_free(pwd);
		ftperrno = FTP_LOSTCONN;
		return NULL;
	}

	if (ftpGetReply(site)[0] != '2') {
		file = ftpFileNew(name, pwd, 0, NULL, 'I', NULL, NULL, NULL);
		if (!file) {
			g_free(pwd);
			ftperrno = FTP_SYS;
			return NULL;
		}
		return file;
	}

	if (!ftpType(site, 'A')) {
		g_free(pwd);
		fclose(site->data);
		return NULL;
	}

	if (!ftpOpenData(site, "rt")) {
		g_free(pwd);
		return NULL;
	}

	if (!ftpCommand(site, "LIST %s %s\r\n", LIST_ARGS, name)) {
		g_free(pwd);
		fclose(site->data);
		ftperrno = FTP_LOSTCONN;
		return NULL;
	}

	if (ftpGetReply(site)[0] != '1') {
		g_free(pwd);
		fclose(site->data);
		ftperrno = FTP_RLISTFAIL;
		return NULL;
	}

	buf = g_malloc(LENGTH);
	if (!buf) {
		g_free(pwd);
		fclose(site->data);
		ftperrno = FTP_SYS;
		return NULL;
	}

	fgets(buf, LENGTH, site->data);
	fclose(site->data);

	ftpGetReply(site);

	if (!ftpChdir(site, site->pwd)) {
		g_free(pwd);
		g_free(buf);
		return NULL;
	}

	file = ParseLs(buf, &site->filelist, dir);
	g_free(pwd);
	g_free(buf);

	if (!queuefile)
		return file;

	/* If we got here, that means this File was created by ftpQueueLoad,
	   thus the date, mode, user, and group fields were not set, so
	   let's set them. */

	queuefile->date = g_strdup(file->date);
	queuefile->mode = g_strdup(file->mode);
	queuefile->user = g_strdup(file->user);
	queuefile->group = g_strdup(file->group);
	queuefile->size = file->size; /* in case the file size has changed */

	site->filelist = g_list_remove(site->filelist, file);
	ftpFileDelete(file);

	return queuefile;
}

ftpFile *
ftpGetFileInfoLocal(const gchar *dir, const gchar *name)
{
	gchar *lscommand, *pwd, *buf;
	GList *filelist;
	ftpFile *file, *queuefile = NULL;
	FILE *ls;
	gint fd, r;

	g_return_val_if_fail(name != NULL, NULL);

	pwd = g_get_current_dir();

	if (!dir)
		dir = pwd;

	r = pthread_mutex_lock(&localfilemutex);
	if (r != 0) {
		ftperrno = FTP_SYS;
		errno = r;
		return NULL;
	}

	for (filelist = localfilelist; filelist; filelist = filelist->next) {
		file = filelist->data;
		if (strcmp(file->dir, dir) == 0
		    && strcmp(file->name, name) == 0) {
			if (!file->date || !file->mode || !file->user
			    || !file->group)
				queuefile = file;
			else {
				g_free(pwd);
				pthread_mutex_unlock(&localfilemutex);
				return file;
			}
		}
	}
	pthread_mutex_unlock(&localfilemutex);

	if (chdir(dir) < 0) {
		g_free(pwd);
		ftperrno = FTP_LCHDIRFAIL;
		return NULL;
	}

	r = ftpLocalFileExists(name);
	if (r < 0) {
		g_free(pwd);
		ftperrno = FTP_SYS;
		return NULL;
	} else if (!r) {
		fd = creat(name, S_IRUSR | S_IWUSR | S_IRGRP);
		if (fd < 0) {
			g_free(pwd);
			ftperrno = FTP_SYS;
			return NULL;
		}
		close(fd);
	}

	lscommand = g_strdup_printf("/bin/ls -alL \"%s\"", name);
	ls = popen(lscommand, "r");
	g_free(lscommand);

	if (!ls) {
		g_free(pwd);
		ftperrno = FTP_LLISTFAIL;
		return NULL;
	}

	buf = g_malloc(LENGTH);

	fgets(buf, LENGTH, ls);
	pclose(ls);

	r = pthread_mutex_lock(&localfilemutex);
	if (r != 0) {
		ftperrno = FTP_SYS;
		errno = r;
		return NULL;
	}

	file = ParseLs(buf, &localfilelist, dir);
	pthread_mutex_unlock(&localfilemutex);
	g_free(pwd);
	g_free(buf);

	if (!queuefile)
		return file;

	/* If we got here, that means this File was created by ftpQueueLoad,
	   thus the date, mode, user, and group fields were not set, so
	   let's set them. */

	queuefile->date = g_strdup(file->date);
	queuefile->mode = g_strdup(file->mode);
	queuefile->user = g_strdup(file->user);
	queuefile->group = g_strdup(file->group);
	queuefile->size = file->size; /* in case the file size has changed */

	r = pthread_mutex_lock(&localfilemutex);
	if (r != 0) {
		ftperrno = FTP_SYS;
		errno = r;
		return NULL;
	}

	localfilelist = g_list_remove(localfilelist, file);
	pthread_mutex_unlock(&localfilemutex);
	ftpFileDelete(file);

	return queuefile;
}

static gint
CompareDir(gconstpointer a, gconstpointer b)
{
	const ftpDirectory *directory;
	const gchar *name;

	directory = a;
	name = b;

	return strcmp(directory->name, name);
}

ftpDirectory *
ftpGetDirRemote(ftpSite *site, const gchar *dir)
{
	ftpDirectory *directory;
	ftpFile *file;
	GList *listitem;
	gchar *buf, *olddir = NULL;

	g_return_val_if_fail(site != NULL, NULL);

	if (dir) {
		if (!ftpChdir(site, dir)) {
			ftperrno = FTP_RCHDIRFAIL;
			g_free(olddir);
			return NULL;
		}
	} else
		dir = site->pwd;

	if ((listitem = g_list_find_custom(site->dirlist, (gchar *) dir,
					   CompareDir))) {
		g_free(olddir);
		return listitem->data;
	}

	if (!ftpOpenData(site, "rt")) {
		g_free(olddir);
		return NULL;
	}

	if (!ftpType(site, 'A')) {
		g_free(olddir);
		return NULL;
	}

	directory = ftpDirectoryNew(dir);
	if (!directory) {
		g_free(olddir);
		return NULL;
	}

	if (!ftpCommand(site, "LIST %s\r\n", LIST_ARGS)) {
		g_free(olddir);
		ftperrno = FTP_LOSTCONN;
		return NULL;
	}

	if (ftpGetReply(site)[0] != '1') {
		g_free(olddir);
		ftperrno = FTP_RLISTFAIL;
		return NULL;
	}

	buf = g_malloc(LENGTH);
	if (!buf) {
		g_free(olddir);
		ftperrno = FTP_SYS;
		return NULL;
	}

	/* eat the "total" line */
	fgets(buf, LENGTH, site->data);

	while (fgets(buf, LENGTH, site->data)) {
		file = ParseLs(buf, &site->filelist, dir);
		if (!file)
			continue;

		if (file->type == 'd')
			directory->dirlist = g_list_append(directory->dirlist,
							   file);
		else
			directory->filelist =
				g_list_append(directory->filelist, file);
	}

	fclose(site->data);
	ftpGetReply(site);

	site->dirlist = g_list_append(site->dirlist, directory);

	if (olddir) {
		ftpChdir(site, olddir);
		g_free(olddir);
	}

	g_free(buf);
	return directory;
}

ftpDirectory *
ftpGetDirLocal(const gchar *dir)
{
	FILE *ls;
	ftpDirectory *directory = NULL;
	ftpFile *file;
	GList *listitem;
	gchar *pwd, *buf = NULL;
	gchar *lscommand;
	gint r;

	pwd = g_get_current_dir();
	if (!dir)
		dir = pwd;

	r = pthread_mutex_lock(&localdirmutex);
	if (r != 0) {
		ftperrno = FTP_SYS;
		errno = r;
		goto out;
	}

	if ((listitem = g_list_find_custom(localdirlist, (gchar *) dir,
					   CompareDir))) {
		directory = listitem->data;
		pthread_mutex_unlock(&localdirmutex);
		goto out;
	}
	pthread_mutex_unlock(&localdirmutex);

	directory = ftpDirectoryNew(dir);
	if (!directory)
		goto out;

	lscommand = g_strdup_printf("/bin/ls -alL %s", dir);
	ls = popen(lscommand, "r");
	g_free(lscommand);

	if (!ls) {
		ftperrno = FTP_LLISTFAIL;
		goto out;
	}

	buf = g_malloc(LENGTH);
	if (!buf) {
		ftperrno = FTP_SYS;
		goto out;
	}

	/* eat the "total" line */
	fgets(buf, LENGTH, ls);

	while (fgets(buf, LENGTH, ls)) {
		r = pthread_mutex_lock(&localfilemutex);
		if (r != 0) {
			ftperrno = FTP_SYS;
			errno = r;
			goto out;
		}

		file = ParseLs(buf, &localfilelist, dir);
		pthread_mutex_unlock(&localfilemutex);
		if (!file)
			continue;

		if (file->type == 'd')
			directory->dirlist = g_list_append(directory->dirlist,
							   file);
		else
			directory->filelist =
				g_list_append(directory->filelist, file);
	}

	pclose(ls);

	r = pthread_mutex_lock(&localdirmutex);
	if (r != 0) {
		ftperrno = FTP_SYS;
		errno = r;
		goto out;
	}

	localdirlist = g_list_append(localdirlist, directory);
	pthread_mutex_unlock(&localdirmutex);

  out:
	g_free(pwd);

	g_free(buf);
	return directory;
}

static ftpFile *
RefreshParseLs(ftpSite *site, gchar *buf, gboolean local, const gchar *dir,
	       GList *dirlist)
{
	gchar *size, *month, *day, *year, *date, *fn, *tmp;
	gchar *mode, *user, *group;
	GList *filelist;
	ftpFile *file = NULL, *testfile;
	gboolean islink = FALSE;

	mode = strtok(buf, " ");
	strtok(NULL, " ");
	user = strtok(NULL, " ");
	group = strtok(NULL, " ");
	size = strtok(NULL, " ");
	month = strtok(NULL, " ");
	day = strtok(NULL, " ");
	year = strtok(NULL, " ");

	/* we can't use strtok because filenames may have spaces */
	fn = year + strlen(year) + 1;
	g_strchomp(fn);
	if (strcmp(fn, ".") == 0 || strcmp(fn, "..") == 0)
		return NULL;

	if (buf[0] == 'l') {
		if ((tmp = strstr(fn, "->"))) {
			islink = TRUE;
			(tmp - 1)[0] = '\0';
		}
	}

	date = g_strdup_printf("%s %s %s", month, day, year);

	for (filelist = dirlist; filelist; filelist = filelist->next) {
		testfile = filelist->data;
		if (strcmp(testfile->dir, dir) == 0
		    && strcmp(testfile->name, fn) == 0) {
			file = testfile;
			break;
		}
	}

	/* FIXME: I don't know what I was thinking; this won't work at all.
	   We're in the middle of a data transfer; this would cause a hang.
	   Need to do ftpFileNew(), and all the other stuff needs to be in
	   an else. */
	if (!file)
		file = (local ? ftpGetFileInfoLocal(dir, fn)
			: ftpGetFileInfoRemote(site, dir, fn));

	g_free(file->name);
	g_free(file->dir);
	g_free(file->date);
	g_free(file->mode);
	g_free(file->user);
	g_free(file->group);

	file->name = g_strdup(fn);
	file->dir = g_strdup(dir);
	file->size = strtol(size, NULL, 10);
	file->date = g_strdup(date);
	file->mode = g_strdup(mode);
	file->user = g_strdup(user);
	file->group = g_strdup(group);

	g_free(date);

	if (buf[0] == 'd')
		file->type = 'd';
	else {
		if (islink)
			file->type = 'l';
		else
			/* eventually we'll determine its type */
			file->type = 'I';
	}

	return file;
}

gboolean
ftpRefreshRemote(ftpSite *site, gchar *dir)
{
	gchar *buf, *oldpwd = NULL;
	GList *olddirlist;
	ftpFile *file;
	ftpDirectory *directory;

	if (dir) {
		if (!ftpChdir(site, dir)) {
			ftperrno = FTP_RCHDIRFAIL;
			g_free(oldpwd);
			return FALSE;
		}
	} else
		dir = site->pwd;

	if (!dir) {
		dir = site->pwd;
	} else {
		if (!ftpChdir(site, dir)) {
			g_free(oldpwd);
			return FALSE;
		}
	}

	directory = ftpGetDirRemote(site, dir);
	olddirlist = directory->dirlist;
	directory->dirlist = NULL;
	g_list_free(directory->filelist);
	directory->filelist = NULL;

	if (!ftpType(site, 'A')) {
		g_free(oldpwd);
		fclose(site->data);
		return FALSE;
	}

	if (!ftpOpenData(site, "rt")) {
		g_free(oldpwd);
		return FALSE;
	}

	if (!ftpCommand(site, "LIST %s\r\n", LIST_ARGS)) {
		ftperrno = FTP_LOSTCONN;
		g_free(oldpwd);
		return FALSE;
	}

	if (ftpGetReply(site)[0] != '1') {
		ftperrno = FTP_RLISTFAIL;
		g_free(oldpwd);
		return FALSE;
	}

	buf = g_malloc(LENGTH);

	/* eat the "total" line */
	fgets(buf, LENGTH, site->data);

	while (fgets(buf, LENGTH, site->data)) {
		file = RefreshParseLs(site, buf, FALSE, dir, olddirlist);
		if (!file)
			continue;

		if (file->type == 'd')
			directory->dirlist = g_list_append(directory->dirlist,
							   file);
		else
			directory->filelist =
				g_list_append(directory->filelist, file);
	}

	fclose(site->data);
	ftpGetReply(site);

	g_free(buf);

	if (oldpwd) {
		ftpChdir(site, oldpwd);
		g_free(oldpwd);
	}

	return TRUE;
}

gboolean
ftpRefreshLocal(gchar *dir)
{
	gchar *buf, *oldpwd, *lscommand;
	FILE *ls;
	GList *olddirlist;
	ftpFile *file;
	ftpDirectory *directory;

	oldpwd = g_get_current_dir();
	if (!dir) {
		dir = oldpwd;
	} else {
		if (chdir(dir) < 0) {
			ftperrno = FTP_LCHDIRFAIL;
			g_free(oldpwd);
			return FALSE;
		}
	}

	directory = ftpGetDirLocal(dir);
	olddirlist = directory->dirlist;
	directory->dirlist = NULL;
	g_list_free(directory->filelist);
	directory->filelist = NULL;

	lscommand = g_strdup_printf("/bin/ls -alL \"%s\"", dir);
	ls = popen(lscommand, "r");
	g_free(lscommand);

	if (!ls) {
		ftperrno = FTP_LLISTFAIL;
		g_free(oldpwd);
		return FALSE;
	}

	buf = g_malloc(LENGTH);

	/* eat the "total" line */
	fgets(buf, LENGTH, ls);

	while (fgets(buf, LENGTH, ls)) {
		file = RefreshParseLs(NULL, buf, TRUE, dir, olddirlist);
		if (!file)
			continue;

		if (file->type == 'd')
			directory->dirlist = g_list_append(directory->dirlist,
							   file);
		else
			directory->filelist =
				g_list_append(directory->filelist, file);
	}

	pclose(ls);

	g_free(buf);
	g_free(oldpwd);

	return TRUE;
}

gchar *
ftp_strerror(ftpError errnum)
{
	g_return_val_if_fail((errnum >= FTP_LOSTCONN) && (errnum < ftp_nerr),
			     NULL);

	if (errnum == FTP_SYS)
		return g_strdup_printf("%s: %s", ftp_errlist[errnum],
				       strerror(errno));

	return g_strdup(ftp_errlist[errnum]);
}

void
ftp_perror(gchar *s)
{
	if (s)
		fprintf(stderr, "%s: ", s);

	if (ftperrno == FTP_SYS)
		perror(ftp_errlist[ftperrno]);
	else
		fprintf(stderr, "%s\n", ftp_errlist[ftperrno]);
}
