/* queue.c
    Routines for manipulating the transfer queue.
    Copyright (C) 1999 Eric Gillespie, Jr.

    This file is part of dpsftplib.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,  as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "dpsftplib.h"
#include "conf.h"
#include "ftpfile.h"
#include "ftpsite.h"
#include "queue.h"

static GList *Queue;
static gboolean QueueGoing;

static ftpQueueItem *
ftpQueueItemNew(ftpSite *site, ftpFile *remotefile, ftpFile *localfile,
		ftpDirection direction)
{
	ftpQueueItem *item;

	g_return_val_if_fail(site != NULL, NULL);
	g_return_val_if_fail(remotefile != NULL, NULL);
	g_return_val_if_fail(localfile != NULL, NULL);
	g_return_val_if_fail((direction == UP) || (direction == DOWN), NULL);

	ftperrno = FTP_SYS;

	item = g_malloc(sizeof(ftpQueueItem));
	if (!item)
		return NULL;

	item->site = ftpSiteNew(site->hostname, site->port, site->login,
				site->pass, site->logfunc, site->command,
				site->reply, site->logdata, site->type);
	if (!item->site) {
		g_free(item);
		return NULL;
	}

	item->remotefile = remotefile;
	item->localfile = localfile;
	item->direction = direction;

	return item;
}

static void
ftpQueueItemDelete(ftpQueueItem *item)
{
	g_return_if_fail(item != NULL);

	if (item->site)
		ftpSiteDelete(item->site);

	g_free(item);
}

static void
AddItem(GScanner *scanner, ftpDirection direction, gchar *url, gchar *local,
	guint32 remotesize, guint32 localsize)
{
	gint i, port;
	HTURI *uri;
	gchar *freehost, *host, *path, *login, *pass, *tmp, *dir;
	gboolean error = FALSE;
	ftpSite *site;
	ftpFile *remotefile, *localfile;

	if (HTCleanTelnetString(url)) {
		g_scanner_error(scanner, "Invalid characters in URL");
		error = TRUE;
	}

	HTSimplify(&url);

	uri = g_malloc(sizeof(HTURI));
	HTScan(url, uri);

	if ((strcmp(uri->access, "http") == 0)) {
		g_scanner_error(scanner, "HTTP URLs not yet supported");
		error = TRUE;
	} else if ((strcmp(uri->access, "ftp") != 0)) {
		g_scanner_error(scanner, "Protocol %s is not supported",
				uri->access);
		error = TRUE;
	}

	if (error) {
		g_free(uri);
		g_scanner_error(scanner, "Item not added to queue1");
		return;
	}

	host = g_strdup(uri->host);
	if (uri->absolute)
		path = g_strdup_printf("/%s", uri->absolute);
	else
		path = g_strdup("/");
	g_free(uri);

	/* The contents of host will change, so save it in freehost to
	   be freed later. */
	freehost = host;
	if (strchr(host, '@')) {
		login = host;
		for (i = 0; login[i] != ':' && i < strlen(login); i++);
		login[i] = '\0';

		pass = login + strlen(login) + 1;
		for (i = 0; pass[i] != '@' && i < strlen(pass); i++);
		pass[i] = '\0';

		host = pass + strlen(pass) + 1;
	} else {
		login = "anonymous";
		pass = Conf.anonpass;
	}

	tmp = strchr(host, ':');
	if (tmp) {
		tmp[0] = '\0';
		tmp++;
		port = atoi(tmp);
	} else
		port = 21;

	site = ftpSiteNew(host, port, login, pass, NULL, NULL, NULL, NULL,
			  'I');
	g_free(freehost);

	dir = g_dirname(path);
	remotefile = ftpFileNew(g_basename(path), dir, remotesize, NULL, 'I',
				NULL, NULL, NULL);
	g_free(dir);

	dir = g_dirname(local);
	localfile = ftpFileNew(g_basename(local), dir, localsize, NULL, 'I',
			       NULL, NULL, NULL);
	g_free(dir);

	if (!ftpQueueAdd(site, remotefile, localfile, direction))
		g_scanner_error(scanner, "Not enough memory to add item");

	g_free(path);
}

static gboolean
GetItem(GScanner *scanner)
{
	GTokenType token;
	ftpDirection direction;
	gchar *url, *local;
	guint32 remotesize, localsize;
	gint pos;

	g_return_val_if_fail(scanner != NULL, FALSE);

	direction = (ftpDirection) scanner->value.v_symbol;

	pos = g_scanner_cur_position(scanner) + 1;
	token = g_scanner_get_next_token(scanner);
	if (token != G_TOKEN_STRING) {
		g_scanner_error(scanner, "String expected at column %d",
				pos);
		return FALSE;
	}
	url = g_strdup(scanner->value.v_string);

	pos = g_scanner_cur_position(scanner) + 1;
	token = g_scanner_get_next_token(scanner);
	if (token != G_TOKEN_STRING) {
		g_scanner_error(scanner, "String expected at column %d",
				pos);
		g_free(url);
		return FALSE;
	}
	local = g_strdup(scanner->value.v_string);

	pos = g_scanner_cur_position(scanner) + 1;
	token = g_scanner_get_next_token(scanner);
	if (token != G_TOKEN_INT) {
		g_scanner_error(scanner, "Integer expected at column %d",
				pos);
		g_free(url);
		g_free(local);
		return FALSE;
	}
	remotesize = scanner->value.v_int;

	pos = g_scanner_cur_position(scanner) + 1;
	token = g_scanner_get_next_token(scanner);
	if (token != G_TOKEN_INT) {
		g_scanner_error(scanner, "Integer expected at column %d",
				pos);
		g_free(url);
		g_free(local);
		return FALSE;
	}
	localsize = scanner->value.v_int;

	AddItem(scanner, direction, url, local, remotesize, localsize);

	g_free(url);
	g_free(local);

	return TRUE;
}

gboolean
ftpQueueLoad(gchar *fn)
{
	GScanner *scanner;
	GTokenType token;
	gint fd, pos;
	gboolean error;

	fd = open(Conf.queuefile, O_RDONLY);
	if (fd < 0)
		return FALSE;

	scanner = g_scanner_new(NULL);
	if (!scanner) {
		close(fd);
		return FALSE;
	}
	g_scanner_input_file(scanner, fd);
	scanner->input_name = Conf.queuefile;

	g_scanner_set_scope(scanner, 42);
	g_scanner_freeze_symbol_table(scanner);
	g_scanner_scope_add_symbol(scanner, 42, "DOWN", (gpointer) DOWN);
	g_scanner_scope_add_symbol(scanner, 42, "UP", (gpointer) UP);
	g_scanner_thaw_symbol_table(scanner);

	while (g_scanner_peek_next_token(scanner) != G_TOKEN_EOF) {
		pos = g_scanner_cur_position(scanner) + 1;
		token = g_scanner_get_next_token(scanner);
		error = FALSE;
		if (token == G_TOKEN_SYMBOL) {
			if (!GetItem(scanner))
				error = TRUE;
		} else {
			g_scanner_error(scanner,
					"UP or DOWN expected at column %d",
					pos);
			error = TRUE;
		}
		if (error) {
			g_scanner_destroy(scanner);
			close(fd);
			return FALSE;
		}
	}

	g_scanner_destroy(scanner);
	close(fd);

	return TRUE;
}

static void
SaveItem(ftpQueueItem *item, FILE *file)
{
	gchar *url, *local;

	if (strcmp(item->site->login, "anonymous") == 0)
		url = g_strdup_printf("%s:%d/%s/%s", item->site->hostname,
			       item->site->port, item->remotefile->dir,
			       item->remotefile->name);
	else
		url = g_strdup_printf("%s:%s@%s:%d/%s/%s", item->site->login,
			       item->site->pass, item->site->hostname,
			       item->site->port, item->remotefile->dir,
			       item->remotefile->name);
	ftpEatSlashes(url, strlen(url));

	local = g_strdup_printf("%s/%s", item->localfile->dir,
				item->localfile->name);
	ftpEatSlashes(local, strlen(local));

	fprintf(file, "%s \"ftp://%s\" \"%s\" %d %d\n",
		(item->direction == UP ? "UP" : "DOWN"),
		url, local, item->remotefile->size, item->localfile->size);

	g_free(url);
	g_free(local);
}

gboolean
ftpQueueSave(gchar *fn)
{
	gint fd;
	FILE *file;

	g_return_val_if_fail(fn != NULL, FALSE);

	fd = creat(fn, S_IRUSR | S_IWUSR);
	if (fd < 0)
		return FALSE;

	file = fdopen(fd, "w");
	if (!file) {
		close(fd);
		return FALSE;
	}

	g_list_foreach(Queue, (GFunc) SaveItem, file);

	fclose(file);

	return TRUE;
}

/* This function is used internally by the library. Do not call it. */
gboolean
ftpQueueInit(void)
{
	Queue = NULL;
	QueueGoing = FALSE;

	if (!ftpQueueLoad(Conf.queuefile))
		fprintf(stderr, "Error loading queue from %s\n",
			Conf.queuefile);

        return TRUE;
}

/* This function is used internally by the library. Do not call it. */
gboolean
ftpQueueShutdown(void)
{
	if (!ftpQueueSave(Conf.queuefile))
		return FALSE;

	ftpQueueClear();

	return TRUE;
}

ftpQueueItem *
ftpQueueAdd(ftpSite *site, ftpFile *remotefile, ftpFile *localfile,
	    ftpDirection direction)
{
	ftpQueueItem *item;

	g_return_val_if_fail(site != NULL, NULL);
	g_return_val_if_fail(remotefile != NULL, NULL);
	g_return_val_if_fail(localfile != NULL, NULL);
	g_return_val_if_fail((direction == UP) || (direction == DOWN), NULL);

	item = ftpQueueItemNew(site, remotefile, localfile, direction);
	if (!item)
		return NULL;

	item->resume = (direction == DOWN ? item->localfile->size
			: item->remotefile->size);

	Queue = g_list_append(Queue, item);

	return item;
}

void
ftpQueueRemove(ftpQueueItem *item)
{
	g_return_if_fail(item != NULL);

	Queue = g_list_remove(Queue, item);
	ftpQueueItemDelete(item);
}

gboolean
ftpQueueMoveUp(ftpQueueItem *item)
{
	GList *element;
	gint pos;

	g_return_val_if_fail(item != NULL, FALSE);

	element = g_list_find(Queue, item);
	if (!element) {
		ftperrno = FTP_QNOTQUEUED;
		return FALSE;
	}

	pos = g_list_position(Queue, element);
	if (pos == 0) {
		ftperrno = FTP_QFIRST;
		return FALSE;
	}

	Queue = g_list_remove(Queue, item);
	Queue = g_list_insert(Queue, item, pos - 1);

	return TRUE;
}

gboolean
ftpQueueMoveDown(ftpQueueItem *item)
{
	GList *element;
	gint pos;

	g_return_val_if_fail(item != NULL, FALSE);

	element = g_list_find(Queue, item);
	if (!element) {
		ftperrno = FTP_QNOTQUEUED;
		return FALSE;
	}

	pos = g_list_position(Queue, element);
	if (pos == g_list_position(Queue, g_list_last(Queue))) {
		ftperrno = FTP_QLAST;
		return FALSE;
	}

	Queue = g_list_remove(Queue, item);
	Queue = g_list_insert(Queue, item, pos + 1);

	return TRUE;
}

void
ftpQueueClear(void)
{
	g_list_foreach(Queue, (GFunc) ftpQueueItemDelete, NULL);
	g_list_free(Queue);
	Queue = NULL;
}

void
ftpQueueStop(void)
{
	QueueGoing = FALSE;
}

gboolean
ftpQueueGoing(void)
{
	return QueueGoing;
}

gint
ftpQueueGetSize(void)
{
	return g_list_length(Queue);
}

const ftpQueueItem *
ftpQueueGetNthItem(gint n)
{
	return g_list_nth_data(Queue, n);
}

const ftpSite *
ftpQueueItemGetSite(const ftpQueueItem *item)
{
	g_return_val_if_fail(item != NULL, NULL);

	return item->site;
}

void
ftpQueueItemSetSite(ftpQueueItem *item, ftpSite *site)
{
	g_return_if_fail(item != NULL);
	g_return_if_fail(site != NULL);

	ftpSiteDelete(item->site);
	item->site = site;
}

const ftpFile *
ftpQueueItemGetRemoteFile(const ftpQueueItem *item)
{
	g_return_val_if_fail(item != NULL, NULL);

	return item->remotefile;
}

void
ftpQueueItemSetRemoteFile(ftpQueueItem *item, ftpFile *file)
{
	g_return_if_fail(item != NULL);
	g_return_if_fail(file != NULL);

	ftpFileDelete(item->remotefile);
	item->remotefile = file;
}

const ftpFile *
ftpQueueItemGetLocalFile(const ftpQueueItem *item)
{
	g_return_val_if_fail(item != NULL, NULL);

	return item->localfile;
}

void
ftpQueueItemSetLocalFile(ftpQueueItem *item, ftpFile *file)
{
	g_return_if_fail(item != NULL);
	g_return_if_fail(file != NULL);

	ftpFileDelete(item->localfile);
	item->remotefile = file;
}

guint32
ftpQueueItemGetResume(const ftpQueueItem *item)
{
	g_return_val_if_fail(item != NULL, 0);

	return item->resume;
}

void
ftpQueueItemSetResume(ftpQueueItem *item, guint32 resume)
{
	g_return_if_fail(item != NULL);

	item->resume = resume;
}

ftpDirection
ftpQueueItemGetDirection(const ftpQueueItem *item)
{
	g_return_val_if_fail(item != NULL, -1);

	return item->direction;
}

const ftpQueueItem *
ftpQueueDataGetItem(const ftpQueueData *data)
{
	g_return_val_if_fail(data != NULL, NULL);

	return data->item;
}

guint32
ftpQueueDataGetSize(const ftpQueueData *data)
{
	g_return_val_if_fail(data != NULL, 0);

	return data->size;
}

guint32
ftpQueueDataGetTransferred(const ftpQueueData *data)
{
	g_return_val_if_fail(data != NULL, 0);

	return data->transferred;
}

ftpQueueStatus
ftpQueueDataGetStatus(const ftpQueueData *data)
{
	g_return_val_if_fail(data != NULL, -1);

	return data->status;
}

static gboolean
TransferItem(ftpQueueItem *item, ftpQueueCallback callback, gpointer userdata,
	     ftpSite *current)
{
	ftpQueueData data;
	ftpSite *site;
	FILE *local = NULL;
	gchar *buf, *fn;
	gint b, read;
	gboolean completed = FALSE;

	g_return_val_if_fail(item != NULL, FALSE);

	buf = g_malloc(Conf.queuebytes);

	data.item = item;
	data.transferred = item->resume;
	
	site = item->site;
	if (Conf.reuseConnection && current)
		if ((current->port == item->site->port)
		    && (strcmp(current->hostname, item->site->hostname) == 0)
		    && (strcmp(current->login, item->site->login) == 0))
			site = current;

	while (!ftpType(site, ftpFileGetType(item->remotefile))) {
		if ((ftperrno == FTP_LOSTCONN)
		    || (ftperrno == FTP_NOTCONNECTED)) {
			ftpClose(site);
			while (!ftpOpen(site)) {
				data.status = FTP_QSTAT_CONNECT_RETRY;
				if (callback)
					callback(&data, userdata);
				if (!QueueGoing)
					goto out;
			}
		}
	}

	data.size = ftpStartTransfer(site, item->remotefile, item->localfile,
				     item->resume, item->direction);
	if (!data.size)
		goto out;

	data.status = FTP_QSTAT_BEGINNING;
	if (callback)
		callback(&data, userdata);

	if (!QueueGoing)
		goto out;

	fn = g_strdup_printf("%s/%s", item->localfile->dir,
			     item->localfile->name);
	data.status = FTP_QSTAT_TRANSFERRING;
	if (item->direction == DOWN) {
		local = fopen(fn, (item->resume ? "a" : "w"));
		do {
			read = ftpRead(site, buf, Conf.queuebytes);
			b = fwrite(buf, 1, read, local);
			if (b > read) {
				/* This should *NOT* happen. */
				goto out;
			}
			data.transferred += b;
			if (b < read) {
				goto out;
			}
			if (callback)
				callback(&data, userdata);
			if (!QueueGoing) {
				goto out;
			}
		} while (data.transferred < data.size);
		fclose(local);
	} else {
		local = fopen(fn, "r");
		while ((read = fread(buf, 1, Conf.queuebytes, local))) {
			b = ftpWrite(site, buf, read);
			if (b > read) {
				/* This should *NOT* happen. */
				goto out;
			}
			data.transferred += b;
			if (!b) {
				if (ftperrno == FTP_TLOSTCONN) {
					goto out;
				}
			} else if (b < read)
				fseek(local, read - b, SEEK_CUR);
			if (callback)
				callback(&data, userdata);
			if (!QueueGoing) {
				goto out;
			}
		}
		fclose(local);
	}

	ftpEndTransfer(site);

	if (callback) {
		data.status = FTP_QSTAT_COMPLETED;
		callback(&data, userdata);
	}

	completed = TRUE;

  out:

	if (!completed) {
		fclose(local);
		ftpAbort(site);
		ftpEndTransfer(site);

		item->resume = data.transferred;

		if (callback) {
			data.status = FTP_QSTAT_ABORTED;
			callback(&data, userdata);
		}
	}

	if (site != current)
		ftpClose(site);

	g_free(buf);

	return completed;
}

/* returns TRUE on successful completion of queue,
   FALSE if ftpQueueStop() was called or an error occurred. */
gboolean
ftpQueueGo(ftpQueueCallback callback, gpointer userdata,
	   ftpLogCallback logfunc, gpointer command, gpointer reply,
	   gpointer logdata,  ftpSite *current)
{
	ftpQueueItem *item;

	g_return_val_if_fail(!QueueGoing, FALSE);

	QueueGoing = TRUE;
	while (g_list_length(Queue) >= 1) {
		item = Queue->data;
		if (logfunc)
			ftpSiteSetLog(item->site, logfunc, command, reply,
				      logdata);
		while (!TransferItem(item, callback, userdata, current))
			if (!QueueGoing)
				return FALSE;
		ftpQueueRemove(item);
	}

	return TRUE;
}
