/* transfer.c
    Routines for transferring files to and from an FTP server.
    Copyright (C) 1999 Eric Gillespie, Jr.

    This file is part of dpsftplib.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2,  as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <arpa/telnet.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>

#include "dpsftplib.h"
#include "ftpfile.h"
#include "ftpsite.h"

struct ftptransfer {
	guint32 total;		/* total number of bytes to be transferred */
	guint32 transferred;	/* bytes transferred so far */
	ftpDirection direction;
	ftpFile *local;
	ftpFile *remote;
};

static ftpTransfer *
ftpTransferNew(guint32 total, guint32 transferred, ftpDirection direction,
	       ftpFile *local, ftpFile *remote)
{
	ftpTransfer *t;

	g_return_val_if_fail((direction == UP) || (direction == DOWN), NULL);

	ftperrno = FTP_SYS;

	t = g_malloc(sizeof(ftpTransfer));
	if (!t)
		return NULL;

	t->total = total;
	t->transferred = transferred;
	t->direction = direction;

	t->local = local;
	t->remote = remote;

	return t;
}

static void
ftpTransferDelete(ftpTransfer *t)
{
	g_return_if_fail(t != NULL);

	g_free(t);
}

guint32
ftpStartTransfer(ftpSite *site, ftpFile *remote, ftpFile *local,
		 guint32 resume, ftpDirection direction)
{
	gchar *pathname = NULL, *mode;
	guint32 size = 0;
	gint r = 0;

	g_return_val_if_fail(site != NULL, 0);
	g_return_val_if_fail(remote != NULL, 0);
	g_return_val_if_fail(local != NULL, 0);
	g_return_val_if_fail((direction == UP) || (direction == DOWN), 0);

	if (!site->connected) {
		ftperrno = FTP_NOTCONNECTED;
		goto out;
	}
	if (site->transfer) {
		ftperrno = FTP_TRANSFERRING;
		goto out;
	}

	pathname = g_strdup_printf("%s/%s", remote->dir, remote->name);
	ftpEatSlashes(pathname, strlen(pathname));

	if (direction == DOWN) {
		if (!ftpCommand(site, "SIZE %s\r\n", pathname)) {
			ftperrno = FTP_LOSTCONN;
			goto out;
		}
		size = strtol(ftpGetReply(site) + 4, NULL, 10);
	} else
		size = local->size;

	mode = (direction == DOWN ? "r" : "w");
	if (!ftpOpenData(site, mode))
		goto out;

	if (resume) {
		if (!ftpCommand(site, "REST %d\r\n", resume)) {
			ftperrno = FTP_LOSTCONN;
			goto out;
		}
		if (ftpGetReply(site)[0] != '3') {
			ftperrno = FTP_RESTFAIL;
			goto out;
		}
	}

	if (direction == DOWN) {
		if (!ftpCommand(site, "RETR %s\r\n", pathname)) {
			ftperrno = FTP_LOSTCONN;
			goto out;
		}
	} else {
		if (!ftpChdir(site, remote->dir))
			goto out;

		if (!ftpCommand(site, "STOR %s\r\n", remote->name)) {
			ftperrno = FTP_LOSTCONN;
			goto out;
		}
	}

	if (ftpGetReply(site)[0] != '1') {
		ftperrno = FTP_TRANSFERFAIL;
		goto out;
	}

	site->transfer = ftpTransferNew(size, resume, direction, local,
					remote);
	if (!site->transfer)
		goto out;

	r = size;

  out:
	if (pathname)
		g_free(pathname);
	return r;
}

gint
ftpRead(ftpSite *site, gchar *buf, gint nbytes)
{
	gint b;

	g_return_val_if_fail(site != NULL, 0);
	g_return_val_if_fail(buf != NULL, 0);

	if (!site->transfer) {
		ftperrno = FTP_NOTTRANSFERRING;
		return 0;
	}

	if (site->transfer->direction != DOWN) {
		ftperrno = FTP_NOTDOWN;
		return 0;
	}

	b = fread(buf, 1, nbytes, site->data);
	if (!b)
		if (ferror(site->data))
			ftperrno = FTP_TLOSTCONN;

	site->transfer->transferred += b;

	return b;
}

gint
ftpWrite(ftpSite *site, gchar *buf, gint nbytes)
{
	gint b;

	g_return_val_if_fail(site != NULL, 0);
	g_return_val_if_fail(buf != NULL, 0);

	if (!site->transfer) {
		ftperrno = FTP_NOTTRANSFERRING;
		return 0;
	}

	if (site->transfer->direction != UP) {
		ftperrno = FTP_NOTUP;
		return 0;
	}

	b = fwrite(buf, 1, nbytes, site->data);
	if (!b)
		if (ferror(site->data))
			ftperrno = FTP_TLOSTCONN;

	site->transfer->transferred += b;

	return b;
}

gboolean
ftpEndTransfer(ftpSite *site)
{
	g_return_val_if_fail(site != NULL, FALSE);

	if (!site->transfer) {
		ftperrno = FTP_NOTTRANSFERRING;
		return FALSE;
	}

	if (site->transfer->direction == DOWN)
		site->transfer->local->size = site->transfer->transferred;
	else
		site->transfer->remote->size = site->transfer->transferred;

	fclose(site->data);
	ftpGetReply(site);
	ftpTransferDelete(site->transfer);
	site->transfer = NULL;

	return TRUE;
}

gboolean
ftpAbort(ftpSite *site)
{
	static const gchar msg[4] = { IAC, IP, IAC };

	g_return_val_if_fail(site != NULL, FALSE);

	if (!site->transfer) {
		ftperrno = FTP_NOTTRANSFERRING;
		return FALSE;
	}

	fflush(site->data);

	if (send(fileno(site->cout), msg, 3, MSG_OOB) != 3) {
		ftperrno = FTP_SYS;
		return FALSE;
	}

	if (!fprintf(site->cout, "%c", DM)) {
		ftperrno = FTP_LOSTCONN;
		return FALSE;
	}

	if (!ftpCommand(site, "ABOR\r\n")) {
		ftperrno = FTP_LOSTCONN;
		return FALSE;
	}

	ftpGetReply(site);

	return TRUE;
}
